# 前端Git规范

[[TOC]]
## Git分支命名

```
master分支: 上线的分支
dev分支: 开发中的分支
dev-xxx: 个人分支, 从dev分支创建, 每日合并到dev
hotfix-xxx: 从master分支创建, 紧急修复上线版本的bug, 修复完后, 需要合并到master和dev分支
bugfix: dev分支中的bug修复. 修复完, 合并到dev分支并删除.
```

git flow 分支参考
* master：主分支，负责记录上线版本的迭代，该分支代码与线上代码是完全一致的。
* dev：开发分支，该分支记录相对稳定的版本，所有的feature分支和bugfix分支都从该分支创建。
* feature/*：特性（功能）分支，用于开发新的功能，不同的功能创建不同的功能分支，功能分支开发完成并自测通过之后，需要合并到 dev 分支，之后删除该分支。
* bugfix/*：bug修复分支，用于修复不紧急的bug，普通bug均需要创建bugfix分支开发，开发完成自测没问题后合并到 dev 分支后，删除该分支。
* release/*：发布分支，用于代码上线准备，该分支从dev分支创建，创建之后由测试同学发布到测试环境进行测试，测试过程中发现bug需要开发人员在该release分支上进行bug修复，所有bug修复完后，在上线之前，需要合并该release分支到master分支和dev分支。
* hotfix/*：紧急bug修复分支，该分支只有在紧急情况下使用，从master分支创建，用于紧急修复线上bug，修复完成后，需要合并该分支到master分支以便上线，同时需要再合并到dev分支。

[阮一峰: Git三种广泛使用的流程](https://www.ruanyifeng.com/blog/2015/12/git-workflow.html)

## tag 

项目测试通过, 上线打包前统一打tag: 描述版本功能、迭代功能和影响范围

`v1.1.1`

第一位代表大版本

第二位代表增加小功能

第三位代表bug修复

## Git Commit Message格式
参考: [Angular 规范](https://github.com/angular/angular.js/blob/master/DEVELOPERS.md#commits)

每次提交，Commit message 都包括三个部分：Header，Body 和 Footer。
```
<type>(<scope>): <subject>
<BLANK LINE>
<body>
<BLANK LINE>
<footer>
```
| type(必需)        | Type of change           | commit类别        |
| ----------------- | ------------------------ | ----------------- |
| **scope(可选)**   | **Scope of this change** | commint影响的范围 |
| subject(必需) | **Short description**    | 简短的描述        |

其中，Header 是必需的，Body 和 Footer 可以省略。

不管是哪一个部分，任何一行都不得超过72个字符（或100个字符）。这是为了避免自动换行影响美观。
### 1.1 type

type用于说明 commit 的类别，只允许使用下面标识

- feat：新功能（feature）
- fix：修补bug
- docs：文档（documentation）
- style： 格式（不影响代码运行的变动,空格,格式化,等等）
- refactor：重构（即不是新增功能，也不是修改bug的代码变动）
- test：增加测试或者修改测试
- chore：构建过程或辅助工具的变动

### 1.2 scope

`scope`用于说明 commit 影响的范围，比如数据层、控制层、视图层等等，视项目不同而不同。
修改影响了不止一处, 用`*`代替
### 1.3 subject

`subject`是 commit 目的的简短描述，不超过50个字符。

```txt
以动词开头，使用第一人称现在时，比如change，而不是changed或changes
第一个字母小写
结尾不加句号（.）
```

### 1.4 body(非必须)
一般采用列表的形式

### 1.5 footer(非必须)
两种情况
(1) 不兼容变动:
如果当前代码与上一个版本不兼容，则 Footer 部分以BREAKING CHANGE开头，后面是对变动的描述、以及变动理由和迁移方法。
(2) 关闭 Issue: 
`closed #xxx`

### 特殊: Revert
如果当前 commit 用于撤销以前的 commit，则必须以revert:开头，后面跟着被撤销 Commit 的 Header。
```
revert: feat(pencil): add 'graphiteWidth' option

This reverts commit 667ecc1654a317a13331b17617d973392f415f02.
```
Body部分的格式是固定的，必须写成This reverts commit &lt;hash>.，其中的hash是被撤销 commit 的 SHA 标识符。

### 相关插件: commitizen

```bash
npm install commitizen -g
# use in npm 
commitizen init cz-conventional-changelog --save-dev --save-exact
# use in yarn 
commitizen init cz-conventional-changelog --yarn --dev --exact
# use 'git cz'
git cz
# 提交信息需要换行, 用/n
```

## 代码合并

使用`git merge --no-ff `合并分支

```bash
# git config --global pull.ff only # Disallows non ff merges on pull. Overrides merge.ff when pulling
git config --global merge.ff false # even create extra merge commit when fast forward merge would be possible
git config --global pull.rebase true # set up pull to rebase instead of merge
```

```bash
# 冲突
1. 手动解决冲突
2. git add
3. git rebase --continue
# 终止rebase
git rebase --abort
```
